/*************************************/
/*      バレーボール用得点表示器     */
/*************************************/
//  2018-01-28
//  バレーボール用
//      SW1     クリア(shift)
//      SW2     マイナス(shift)
//      SW3     Bセット+1
//      SW4     Aセット+1
//      SW5     タイマー
//      SW6     Aサーブ
//      SW7     Bポイント+1
//      SW8     Aポイント+1
//  セット、サーブ表示LED
//      A2      COM0 H seg 左10桁
//      A1      COM1 H     左1桁
//      B2      COM2 H     右10桁
//      B1      COM3 H     右1桁
//      ｻｰﾌﾞA   COM4 H seg
//      ｻｰﾌﾞB   COM4 A seg

/*****  ライブラリ      ****/
#include <MsTimer2.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include  "wiring_private.h"      // sbi,cbi命令用
// http://www.arduino.cc/en/Tutorial/LiquidCrystal

/*****  配列のﾃﾞｰﾀ数を返すﾏｸﾛ    *****/
#define  DIMSIZ(a)    (sizeof(a)/sizeof(*a))
//  nop
#define  NOP       asm("nop")

/*****  SW コード  *****/
//  bit マスクデータ
#define SWM_CLR      (1 << 0)       // 押しながらでクリア
#define SWM_MNS      (1 << 1)       // 押しながらで-1
//  SW on 確定データ
#define SW_APNT     1           // A得点+1
#define SW_APNTM    2           // A得点-1
#define SW_APCLR    3           // A得点クリア
#define SW_ASET     4           // Aセット+1
#define SW_ASETM    5           // Aセット-1
#define SW_ASCLR    6           // Aセットクリア
#define SW_BPNT     7           // B得点+1
#define SW_BPNTM    8           // B得点-1
#define SW_BPCLR    9           // B得点クリア
#define SW_BSET     10          // Bセット+1
#define SW_BSETM    11          // Bセット-1
#define SW_BSCLR    12          // Bセットクリア
#define SW_TIM      13          // タイマー
#define SW_SRV      14          // サーブ 

/*****  グローバルデータ    *****/
//  表示データ
byte disp_bff[5];       // 5ﾊﾞｲﾄ分の表示ﾊﾞｯﾌｧ
                        // com_cntでCOM駆動
                        // 0:得点A 10桁
                        // 1:得点A  1桁
                        // 2:得点B 10桁
                        // 3:得点B  1桁
                        // 4:サーブAB表示
//  COM駆動データ Hでｵﾝ
//  COM0:PD2 ～ COM4:PD6
byte com_cnt;          // 0～4の5ﾋﾞｯﾄｶｳﾝﾄ
byte com_drv[] = {
    0b00000100,        // COM0 PD2  
    0b00001000,        // COM1 PD3
    0b00010000,        // COM2 PD4  
    0b00100000,        // COM3 PD5  
    0b01000000,        // COM4 PD6  
};
//  7seg セグメントデータ 0～9
byte seg_7seg[] = {
    0b00111111,       // 0 HGFEDCBA
    0b00000110,       // 1
    0b01011011,       // 2
    0b01001111,       // 3
    0b01100110,       // 4
    0b01101101,       // 5
    0b01111101,       // 6
    0b00000111,       // 7
    0b01111111,       // 8
    0b01101111,       // 9
};

//  得点、セット、サーブデータ
byte point_a;           // A得点   0～99
byte set_a;             // Aセット 0,1,2
byte point_b;           // B得点   0～99
byte set_b;             // Bセット 0,1,2
byte srv_ab;            // サーブ  0:未決定 1:A 2:B
//  動作モード,タイマーデータ
byte tim_mode;          // 0:得点表示,1:タイマーモード
byte tim_1sec;          // 1秒タイマー(ダウンカウント)
byte tim_1sec10;        // 1秒用10mSカウント
byte f_tim1sec;         // 1秒経過フラグ
//  ブザー報知
byte bzz_exc;           // ブザー報知実行区分
byte bzz_tm;            // 10msでダウンカウント
byte bzz_cnt;           // 報知回数

/**************************/
/*    シリアル出力        */
/**************************/
/*****  書式付シリアル出力     *****/
void txprintf(const char *s, ...)
{
va_list vp;
char bff[40];     // ﾊﾞｯﾌｧを確保
    va_start(vp,s);
    vsprintf(bff,s,vp);
    Serial.print(bff);
    va_end(vp);
}

/**********************************/
/*      セットアップ              */
/**********************************/
/*****  Set Up      *****/
//  ATmega328Pのレジスタを直接制御
void setup()
{
//  I/Oイニシャル
    PORTB = 0b00000000;    // out data, pull up
    DDRB  = 0b00110011;    // portI/O
    //          |||||+----  PB0  out -
    //          ||||+-----  PB1  out PWM OC1A
    //          |||+------  PB2  in  SW0 ｺﾓﾝ入力 pullﾀﾞｳﾝ
    //          ||+-------  PB3  in  SW1 ｺﾓﾝ入力
    //          |+--------  PB4  out SEG G 駆動
    //          +---------  PB5  out SEG H
    PORTC = 0b00000000;    // out data, pull up
    DDRC  = 0b00111111;    // portI/O
    //          |||||+----  PC0  out SEG A 駆動
    //          ||||+-----  PC1  out SEG B
    //          |||+------  PC2  out SEG C
    //          ||+-------  PC3  out SEG D
    //          |+--------  PC4  out SEG E
    //          +---------  PC5  out SEG F
    PORTD = 0b00000001;    // out data, pull up
    DDRD  = 0b11111110;    // portI/O
    //        |||||||+----  PD0  in  RXD
    //        ||||||+-----  PD1  out TXD
    //        |||||+------  PD2  out COM0駆動
    //        ||||+-------  PD3  out COM1駆動
    //        |||+--------  PD4  out COM2駆動
    //        ||+---------  PD5  out COM3駆動
    //        |+----------  PD6  out COM4駆動
    //        +-----------  PD7  out -
//  シリアル
    Serial.begin(9600);     //  シリアル通信
//  タイマー割り込み
    MsTimer2::set(1, int1ms);           // 1ms割り込み
    MsTimer2::start();                  // 割り込み処理開始
//  PWM / timer1 OC1A出力
    TCNT1 = 0;
    TCCR1A = 0b10000000;
    //         ||||  ++----  WGM11,10 CTC OCR1A
    //         ||++--------  COM1B I/Oポート
    //         ++----------  COM1A 一致でL出力
    TCCR1B = 0b00001010;
    //         || ||+++----  CS 16MHz/8 =2MHz 
    //         || ++-------  WGM13,12 PWM ICR1
    //         |+----------  ICES1
    //         +-----------  ICNC1
    OCR1A = 500;            // 2kHz
}

/******************************/
/*      スイッチ入力          */
/******************************/
//  SWスキャンデータ SWは8コ
//  bit 1,0にon(1)/off(0)データが入る
byte sw_scan[5];        // COM0～4でスキャン    
                        // データ位置はcom_cnt
byte sw_data;           // SWのonを1ﾊﾞｲﾄにまとめたﾃﾞｰﾀ
byte sw_chk[6];         // onエッジチェックデータ
byte sw_inp;            // SWのon/off状態
byte sw_on;             // onエッジ 1でon
byte f_swscan;          // SWﾁｪｯｸﾀｲﾐﾝｸﾞ sw_on確定を知らせる

/*****    スイッチ入力チェック      *****/
//  5mSｻｲｸﾙで6回ﾁｪｯｸ, onエッジを見つける
//  タイマー割り込み内で処理    
void swscan(void)
{
static byte p = 0;      // 最新入力書き込み位置
byte i;
byte m, d0, d1;
    sw_chk[p] = sw_data;            // 最新SWﾃﾞｰﾀ Hでon
    d0 = 0x00;                      // Lﾁｪｯｸ
    d1 = 0xFF;                      // Hﾁｪｯｸ (1bit)
    for(i = 0; i < DIMSIZ(sw_chk); i++){    // H,L安定確認
        d0 |= sw_chk[i];
        d1 &= sw_chk[i];
    }
    m = (~d0) | d1;                 // 1のところが安定データ
    d0 = sw_inp;
    sw_inp = (d0 & (~m)) | (sw_chk[p] & m);     // 確定
    sw_on |= (~d0) & sw_inp;        // onエッジをチェック
    p++;                            // 次書き込み位置
    if(p >= DIMSIZ(sw_chk))     p = 0;  // ﾎﾟｲﾝﾀ一周
}

/*****  スイッチ on データ     *****/
//  +1,-1,CLRのチェック
typedef struct{
    byte    pls;
    byte    mns;
    byte    clr;
}st_swcode;
st_swcode sw_code[] = {
        SW_BSET, SW_BSETM, SW_BSCLR,      // Bセット +1,-1,clr
        SW_ASET, SW_ASETM, SW_ASCLR,      // Aセット +1,-1,clr
        SW_SRV,  SW_SRV,   SW_SRV,        // サーブ 
        SW_TIM,  SW_TIM,   SW_TIM,        // タイマー
        SW_BPNT, SW_BPNTM, SW_BPCLR,      // B得点 +1,-1,clr
        SW_APNT, SW_APNTM, SW_APCLR,      // A得点 +1,-1,clr
};

/*****  SW on チェック      *****/
//  オフなら0、オンならswコードを返す
//  sw_onのbit 1,0はシフト状態なので無視
byte swonchk(void)
{
byte m, i, r;
    r = 0;                              // SW onなし
    m = 0b00000100;                     // bit 2から
    for(i = 0; i < 6 ;i++){             // 6bit
        if(sw_on & m){                  // SW on?
            sw_on &= ~m;
            if(sw_inp & SWM_MNS)       r = sw_code[i].mns;
            else if(sw_inp & SWM_CLR)  r = sw_code[i].clr;
            else                       r = sw_code[i].pls;
            break;
        }
        m <<= 1;
    }
    return r;    
}

/******************************/
/*      1ms タイマー割り込み  */
/******************************/
/*****  タイマーデータ    *****/
volatile byte  f_tm1ms;         // 1ms確定フラグ
volatile byte  f_1sec;          // 1秒確定フラグ
volatile byte  tm_10ms;         // 計時用ダウンカウント
volatile byte  f_10ms;          // 10msフラグ
volatile byte  f_blink;         // 点滅フラグ
                                // 0:0.2秒 1:0.8秒
/*****  割り込み実行      *****/
//  1msサイクル　SEG駆動ｵﾌ　COM駆動 次chに
void int1ms(void)
{
static byte cnt10   = 0;
static word cnt1000 = 0;
static byte tm_blk  = 0;    // 10msで点滅処理
byte i;
//  セグメント駆動オフ
    PORTC &= 0b11000000;        // SEG A～F off
    PORTB &= 0b11001111;        // SEG G,H off
//  SWスキャン(PB2,PB3)
    sw_scan[com_cnt] = (PINB & 0b00001100) >> 2;
//  COM駆動オフ
    PORTD &= 0b10000011;        // COM駆動off
//  待ち(隣接光漏れ対策)
    for(i = 0; i < 20; i++){
        NOP;
    }
//  1mS
    f_tm1ms = 1;                // 1ms確定
//  10msタイマー
    cnt10++;
    if(cnt10 >= 10){                // 10ms
      cnt10 = 0;
      f_10ms = 1;                   // 10msフラグ
      if(tm_10ms)     tm_10ms--;    // 10mS ダウンカウント
      if(tm_blk)      tm_blk--;     // 点滅
      if(tm_blk == 0){
        f_blink ^= 1;               // 点滅状態反転
        if(f_blink)     tm_blk = 90;  // 0.9秒
        else            tm_blk = 10;  // 0.1秒
      }
      tim_1sec10++;                 // 1秒タイマー
      if(tim_1sec10 >= 100){
        tim_1sec10 = 0;
        f_tim1sec  = 1;
        if(tim_1sec)    tim_1sec--;
      }
    }
//  1secタイマー
    cnt1000++;
    if(cnt1000 >= 1000){
      cnt1000 = 0;
      f_1sec = 1;               // フラグ オン
    }
//  次COM駆動 COM0～COM4, ｽｲｯﾁﾁｪｯｸ
    com_cnt++;
    if(com_cnt >= DIMSIZ(com_drv)){     // 1周?
        com_cnt = 0;                    // COM0から
        sw_data = sw_scan[0] |          // SW1,2
                 (sw_scan[2] << 2) |    // SW3,4
                 (sw_scan[4] << 4) |    // SW5,6
                 (sw_scan[1] << 6);     // SW7,8
        swscan();                       // SW onチェック
        f_swscan =1 ;                   // sw_on,sw_inpを確定
    }
    PORTD |= com_drv[com_cnt];      // COM出力ｵﾝ
//  ｾｸﾞﾒﾝﾄ駆動
    PORTC |= disp_bff[com_cnt] & 0x3F;        // SEG F～A
    PORTB |= (disp_bff[com_cnt] & 0xC0) >> 2; // SEG H,G
}

/*****************************/
/*      ブザー出力           */
/*****************************/
//  周波数
word bzz_frq;

/*****  ブザーon       *****/
//  周波数を指定
void bzzon(void)
{
    OCR1A = 1000000 / bzz_frq;      // 1MHz / frq
    TCCR1A = 0b01000000;
    //         ||||  ++----  WGM11,10 CTC OCR1A
    //         ||++--------  COM1B I/Oポート
    //         ++----------  COM1A 一致でトグル
}
/*****  ブザーoff       *****/
void bzzoff(void)
{
    TCCR1A = 0b10000000;
    //         ||||  ++----  WGM11,10 CTC OCR1A
    //         ||++--------  COM1B I/Oポート
    //         ++----------  COM1A 一致でL出力
}

/*****  ブザー実行   *****/
//  10mSサイクルで実行
void bzzexc(void)
{
    if(bzz_tm)      bzz_tm--;     // ブザー報知タイマ
    if(bzz_tm == 0){
      switch(bzz_exc){
        case 1:
          bzzoff();
          if(bzz_cnt & 1){    // 3,1
              bzz_tm = 18;      // 180ms
          }
          else{
              bzz_tm = 6;       // 60ms
          }
          bzz_exc++;
          break;
        case 2:
          if(bzz_cnt)   bzz_cnt--;
          if(bzz_cnt == 0){
            OCR1A = 250;         // 4kHzで
            TCNT1 = 0;
            bzz_exc = 0;
          }
          else{                 // 繰り返し
            bzzon();
            bzz_tm = 6;         // 60ms
            bzz_exc--;
          }
          break;
      }
    }   
}

/*****  Ｂｉ      *****/
void bzzbi(void)
{
    bzz_frq = 120;          // 120Hz
    bzzon();
    bzz_tm = 6;             // 60ms
    bzz_cnt = 1;
    bzz_exc = 1;
}
/*****  Ｐｉ      *****/
void bzzpi(void)
{
    bzz_frq = 2000;         // 2000Hz
    bzzon();
    bzz_tm = 2;             // 20ms
    bzz_cnt = 1;
    bzz_exc = 1;
}
/*****  ＰｉＰｉＰｉＰｉ    *****/
void pipi4(void)
{
    bzz_frq = 2000;            // 2kHz
    bzzon();
    bzz_tm = 6;                // 60ms
    bzz_cnt = 8;
    bzz_exc = 1;
}
/*****  ＰｉＰｉ    *****/
void pipi(void)
{
    bzz_frq = 2000;            // 2kHz
    bzzon();
    bzz_tm = 6;                // 60ms
    bzz_cnt = 2;
    bzz_exc = 1;
}

/******************************/
/*      表示                  */
/******************************/
/*****  得点A、セットA表示   *****/
void dispa(void)
{
byte a, d1, d0;
    d0 = seg_7seg[point_a % 10];     //  0～ 9
    a = point_a / 10;                // 00～90
    if(a == 0)   d1 = 0x00;          // ゼロブランク
    else         d1 = seg_7seg[a];
//  セット A 0,1,2
    switch(set_a){
      case 1:
        d0 |= 0x80;         // seg H
        break;
      case 2:
        d0 |= 0x80;
        d1 |= 0x80;
        break;
    }
    disp_bff[0] = d1;           // A 10桁+1桁
    disp_bff[1] = d0;
}

/*****  得点B、セットB表示   *****/
void dispb(void)
{
byte a, d1, d0;
    d0 = seg_7seg[point_b % 10];     //  0～ 9
    a = point_b / 10;                // 00～90
    if(a == 0)   d1 = 0x00;          // ゼロブランク
    else         d1 = seg_7seg[a];
//  セット B 0,1,2
    switch(set_b){
      case 1:
        d0 |= 0x80;         // seg H
        break;
      case 2:
        d0 |= 0x80;
        d1 |= 0x80;
        break;
    }
    disp_bff[2] = d1;           // B 10桁+1桁
    disp_bff[3] = d0;
}

/*****  サーブ表示       *****/
void dispsrv(void)
{
    switch(srv_ab){
      case 0:       // 未定
        disp_bff[4] &= 0b01111110;      // seg H,seg A off
        break;
      case 1:       // サーブA
        disp_bff[4] |= 0b10000000;      // seg H
        disp_bff[4] &= 0b11111110;
        break;
      case 2:       // サーブB
        disp_bff[4] |= 0b00000001;      // seg A
        disp_bff[4] &= 0b01111111;
        break;
    }
}

/*****  タイマー表示      *****/
void disptim(void)
{
    disp_bff[3] = seg_7seg[tim_1sec % 10];        //  0～  9
    disp_bff[2] = seg_7seg[(tim_1sec / 10) % 6];  //  00～ 59
    disp_bff[1] = seg_7seg[tim_1sec / 60];        // 000～255
    disp_bff[0] = 0;
    if(tim_1sec10 < 50){
        disp_bff[4] |= 0b00010100;   // seg EC点灯
    }
    else{
        disp_bff[4] &= 0b00000000;   // seg ECは消灯
    }
}

/*****  スイッチ入力チェック      *****/
void swinchk(void)
{
byte sw;
    if(f_swscan){           // SW スキャンタイミング
        f_swscan = 0;
        sw = swonchk();
//  モード区分
        if(tim_mode == 0){      //　得点表示モード
          switch(sw){
            case SW_APNT:      // A得点+1
                if(point_a < 99)    point_a++;
                srv_ab = 1;
                bzzbi();
                break;
            case SW_APNTM:     // A得点-1
                if(point_a)      point_a--;
                bzzpi();
                break;
            case SW_ASET:      // Aセット+1
                if(set_a < 2)    set_a++;
                 bzzbi();
               break;
            case SW_ASETM:     // Aセット-1
                if(set_a)        set_a--;
                bzzpi();
                break;
            case SW_BPNT:      // B得点+1
                if(point_b < 99)    point_b++;
                srv_ab = 2;
                bzzbi();            // ブザー Bi
                break;
            case SW_BPNTM:     // B得点-1
                if(point_b)      point_b--;
                bzzpi();
                break;
            case SW_BSET:      // Bセット+1
                if(set_b < 2)    set_b++;
                bzzbi();
                break;
            case SW_BSETM:     // Bセット-1
                if(set_b)        set_b--;
                bzzpi();
                break;
            case SW_APCLR:     // A得点クリア
            case SW_BPCLR:     // B得点クリア
                point_a = point_b = 0;
                bzzpi();
                break;
            case SW_ASCLR:     // Aセットクリア
            case SW_BSCLR:     // Bセットクリア
                set_a = set_b = 0;
                srv_ab = 0;
                bzzpi();
                break;
            case SW_TIM:       // タイマー
                tim_1sec = 30;      // 30秒
                tim_1sec10 = 0;
                f_tim1sec = 0;
                tim_mode = 1;
                pipi();             // PiPi
                break;
            case SW_SRV:       // サーブ 
                srv_ab++;      // 0,1,2
                if(srv_ab > 2)      srv_ab = 1;
                bzzbi();
                break;
            }
         }
//  タイマーモード
         else{
           switch(sw){
             case SW_TIM:       // タイマー
                tim_mode = 0;
                break;
           }
         }
      }      
}

/******************************/
/*      ループ                */
/******************************/
/*****  LOOP      *****/
void loop()
{
    memset(disp_bff, 0xFF, DIMSIZ(disp_bff));   // LED全点灯
    while(f_1sec == 0);                         // 1秒待ち
    memset(disp_bff, 0x00, DIMSIZ(disp_bff));   // 消灯
    sw_on = 0;
//  実行開始
    while(1){
//  スイッチ入力チェック
        swinchk();
//  得点表示
        if(f_10ms){         // 10msサイクルで処理
          f_10ms = 0;
          bzzexc();               // ブザー報知
          if(tim_mode == 0){      //　得点表示モード
              if((f_blink != 0) ||    // 0.9秒 表示中
                 (srv_ab  == 0)){     // サーブ未決定
                dispa();              // A得点,Aセット
                dispb();              // B得点,Bセット
                dispsrv();            // サーブ A,B
              }
              else if(srv_ab == 1){   // サーブA点滅
                disp_bff[0] = disp_bff[1] = 0x00;
                disp_bff[4] &= 0b01111111;       // seg H
              }
              else if(srv_ab == 2){   // サーブB点滅
                disp_bff[2] = disp_bff[3] = 0x00;
                disp_bff[4] &= 0b11111110;       // seg A
              }
              disp_bff[4] &= 0b10101011;   // seg GECは消灯
           }
//  タイマーモード
           else{
              disptim();              // タイマー表示
              if(f_tim1sec){          // 1秒経過  
                f_tim1sec = 0;
                switch(tim_mode){                
                  case 1:           // タイムアップ?
                    if(tim_1sec == 0){
                      pipi4();        // PiPi *4
                      tim_mode++;
                    }
                    break;
                  case 2:           // 報知完了?
                    if(bzz_exc ==0){
                      tim_mode = 0;     // おわり
                    }
                    break;
                }
              }
           }
        }
    }
}

/*====  end of vb_disp1 =====*/

